import os
from dataclasses import dataclass, field
from typing import List, Optional, Any
from fs.base import FS

from zetsubou.commands.execute_stage import execute_stage
from zetsubou.utils import logger
from zetsubou.utils.common import null_or_empty, fix_path
from zetsubou.project.config_matrix import ConfigMatrix
from zetsubou.project.base_context import BaseContext
from zetsubou.project.model.target import Target
from zetsubou.project.model.toolchain import Toolchain
from zetsubou.project.runtime.resolve import ResolvedTarget
from zetsubou.project.runtime.project_loader import ProjectTemplate, resolve_venv
from zetsubou.utils.error_codes import EErrorCode

@dataclass
class Fastbuild:
    emit_fs : FS
    bff_dir : str
    bff_file : str


@dataclass
class Conan:
    yml_files: List[str] = field(default_factory=list)
    dependencies: List[Target] = field(default_factory=list)
    resolved_targets: List[ResolvedTarget] = field(default_factory=list)


@dataclass
class CommandContext(BaseContext):
    command_args: Any = None

    project_template: Optional[ProjectTemplate] = None
    toolchains: List[Toolchain] = field(default_factory=list)
    config_matrix: Optional[ConfigMatrix] = None
    resolved_targets: List[ResolvedTarget] = field(default_factory=list)
    fastbuild: Optional[Fastbuild] = None
    conan: Conan = field(default_factory=Conan)

    def to_out_path(self, path : str):
        return os.path.join(self.fs_root, path)

    def resolve_venv(self):
        if null_or_empty(self.fs_venv):
            venv_path = execute_stage(lambda: resolve_venv(self.project_fs, self.project_template.project, self.fs_root),
                                        'Virtual environemnt found',
                                        EErrorCode.UNABLE_TO_FIND_VENV)

            logger.Info(f"Virtual environement - '{venv_path}'")
            self.fs_venv = fix_path(venv_path)



from typing import List, Optional, Union
from enum import Enum
from dataclasses import dataclass, field

from zetsubou.project.model.filter import TargetFilter
from zetsubou.project.model.target import TargetData
from zetsubou.project.model.toolchain import ECompilerFamily
from bentoudev.dataclass.base import loaded_from_file


class EArch(Enum):
    x64 = 1
    x86 = 2


class ESystem(Enum):
    Windows = 1
    Linux = 2


class EVersionSelector(Enum):
    all = 1,
    latest = 2


@dataclass
class PlatformData(TargetData):
    filter: TargetFilter = field(default_factory=TargetFilter)


@dataclass
class IPlatformToolchain:
    family: ECompilerFamily
    version: Union[EVersionSelector, List[str]]


@dataclass
class IWindowsToolchain(IPlatformToolchain):
    toolset: Union[EVersionSelector, List[str]]


@dataclass
class ILinuxToolchain(IPlatformToolchain):
    libcxx: List[str]


@dataclass
@loaded_from_file
class Platform(TargetData):
    platform: str = None
    host_system: ESystem = None
    host_arch: EArch = None
    target_arch: EArch = None
    filters: Optional[List[PlatformData]] = field(default_factory=list)
    toolchains: List[Union[IPlatformToolchain, IWindowsToolchain, ILinuxToolchain]] = field(default_factory=list)



from zetsubou.utils.subprocess import call_process
import zetsubou.utils.logger as logger
from typing import List


def check_conan_present():
    out, err = call_process(['conan', '--version'], capture=True, realtime=False)
    if err is not None:
        logger.CriticalError('Conan is missing, unable to proceed')
        return False

    logger.Success(out)
    return True


def call_conan(args : List[str], cwd : str):
    verbose = logger.IsVisible(logger.ELogLevel.Verbose)
    out, err = call_process(['conan'] + args, capture = not verbose, realtime = verbose, cwd = cwd)
    if err is not None:
        if not verbose:
            logger.CriticalError(f'Conan command failed: {" ".join(args)}')
            logger.Error(out)

        logger.CriticalError(f"Conan returned error '{err}'")
        return False
    return True



import os
from typing import List

from bentoudev.dataclass.yaml_loader import LineLoader

from zetsubou.commands.command_context import CommandContext
from zetsubou.project.model.platform import EArch, ESystem, Platform
from zetsubou.project.model.toolchain import Toolchain
from zetsubou.utils.common import is_in_enum
from zetsubou.utils.error import ProjectError
from zetsubou.utils.yaml_tools import to_yaml
from zetsubou.utils import logger


def discover_toolchain_list(context: CommandContext) -> List[Toolchain] :
    system_toolchains = discover_toolchain_by_system(context.host_system, context.host_arch)

    # filter by platform
    toolchains = list(filter(lambda t: filter_toolchains_by_target_arch(t, context.project_template.platforms), system_toolchains))

    # filter by whatever is known to Conan
    user_dir = os.path.expanduser('~')
    conan_settings_path = os.path.join(user_dir, '.conan/settings.yml')

    if os.path.exists(conan_settings_path):
        with open(conan_settings_path, encoding='utf-8') as conan_settings_file:
            loader = LineLoader(conan_settings_file.read())
            loaded_yaml = loader.get_single_data()

            if context.host_system == ESystem.Windows:
                toolchains = list(filter(lambda t: filter_toolchains_by_conan_toolchain(t, loaded_yaml), toolchains))

            num_toolchains = len(toolchains)

            if num_toolchains == 0:
                logger.Error("No suitable toolchains found!")

                unsuitable_toolchains = to_yaml(system_toolchains)
                logger.Warning(f"Available toolchains:\n{unsuitable_toolchains}")

                if context.host_system == ESystem.Windows:
                    toolkits = '\n'.join([ name for name in loaded_yaml['compiler']['Visual Studio']['toolset'] ])
                    logger.Warning(f"Available VS toolkits:\n{toolkits}")

                return None
            else:
                logger.Verbose(f"Found ({num_toolchains}) suitable toolchains.")

                if context.command_args.verbose:
                    found_toolchains = to_yaml(toolchains)
                    logger.Verbose(f"Entries:\n{found_toolchains}")

    else:
        logger.Warning('Unable to filter toolchains by Conan settings!')

    return toolchains


def discover_toolchain_by_system(host_system: ESystem, host_arch: EArch) -> List[Toolchain] :
    if host_system == ESystem.Windows:
        from zetsubou.system.windows import windows
        return windows.discover_toolchain_list(host_arch)
    else:
        raise EnvironmentError(f'Sorry, discovering toolchains on \"{host_system.name}\" arch \"{host_arch.name}\" is not currently supported!')


def filter_toolchains_by_target_arch(toolchain:Toolchain, platforms:List[Platform]):
    for plat in platforms:
        if not is_in_enum(toolchain.arch, EArch):
            logger.Warning(f"Unknown arch '{toolchain.arch}' for toolchain '{toolchain.name}'")
            return False

        if EArch[toolchain.arch] == plat.target_arch:
            return True

    return False


def filter_toolchains_by_conan_toolchain(toolchain:Toolchain, conan_settings):
    try:
        if toolchain.Toolset in conan_settings['compiler']['Visual Studio']['toolset']:
            return True
    except Exception:
        return False
    return False



from typing import List
from dataclasses import dataclass, field
from enum import Enum

from zetsubou.project.model.filter import TargetFilter
from zetsubou.project.model.runtime_library import ERuntimeLibrary
from zetsubou.project.model.target import TargetData
from bentoudev.dataclass.base import loaded_from_file


# For compatibility purposes, custom configurations must fallback to one of the default ones, to communicate with VS and Conan
class EBaseConfiguration(Enum):
    DEBUG = 0
    RELEASE = 1
    RELEASE_WITH_DEBUG_INFO = 2


@dataclass
class ConfigurationData(TargetData):
    filter: TargetFilter = field(default_factory=TargetFilter)


@dataclass
@loaded_from_file
class Configuration(TargetData):
    configuration: str = None
    base_configuration: EBaseConfiguration = None
    filters: List[ConfigurationData] = field(default_factory=list)
    runtime_library: ERuntimeLibrary = ERuntimeLibrary.DYNAMIC_RELEASE



from typing import Optional
from dataclasses import dataclass

from zetsubou.project.model.kind import ETargetKind
from zetsubou.project.model.runtime_library import ERuntimeLibrary


@dataclass
class TargetFilter():
    config_string: str = '*'
    kind: Optional[ETargetKind] = None
    runtime_library: Optional[ERuntimeLibrary] = None



from sys import exit as sys_exit
from .zetsubou import main as zet_main
# import cProfile
# import time

# import yaml
# from yaml.composer import Composer
# from yaml.constructor import Constructor

# import ruamel.yaml as ruamel

# def load_content(filename:str):
#     with open(filename, 'r') as file:
#         return file.read()

# class ProfileScope:
#     def __init__(self, print_stats:bool, label:str = ''):
#         self.pr = cProfile.Profile()
#         self.print_stats = print_stats
#         self.label = label

#     def __enter__(self):
#         self.pr.enable()
#         self.start_time = time.time()

#     def __exit__(self, exc_type, exc_val, exc_tb):
#         end_time = time.time()
#         self.pr.disable()

#         print(f'{self.label} took: {end_time - self.start_time}')

#         if self.print_stats:
#             self.pr.print_stats(sort='time')


# def test_ruamel(content:str):
#     with ProfileScope(print_stats=False, label='Ruamel'):
#         ruamel.load(content)


# def test_ruamel_new(content:str):
#     with ProfileScope(print_stats=False, label='Ruamel New'):
#         yml = ruamel.YAML(typ='rt', pure=True)
#         yml.load(content)


# def test_yaml(content:str):
#     with ProfileScope(print_stats=False, label='Yaml'):
#         yaml.load(content)


# def test_yaml_lines(content:str):
#     with ProfileScope(print_stats=False, label='Yaml Lines'):
#         loader = yaml.Loader(content)
#         def compose_node(parent, index):
#             # the line number where the previous token has ended (plus empty lines)
#             line = loader.line
#             node = Composer.compose_node(loader, parent, index)
#             node.__line__ = line + 1
#             return node
#         def construct_mapping(node, deep=False):
#             mapping = Constructor.construct_mapping(loader, node, deep=deep)
#             mapping['__line__'] = node.__line__
#             return mapping
#         loader.compose_node = compose_node
#         loader.construct_mapping = construct_mapping
#         data = loader.get_single_data()


# def profile_yaml_loaders(filename:str):
#     content = load_content(filename)

#     test_yaml_lines(content)
#     test_yaml(content)
#     test_ruamel(content)
#     test_ruamel_new(content)


# profile_yaml_loaders('example/configurations/MsvcRules.yml')

sys_exit(zet_main())



from typing import List, Dict
from zetsubou.project.model.platform import EVersionSelector, IWindowsToolchain
from zetsubou.project.model.toolchain import ECompilerFamily, Toolchain
from zetsubou.project.runtime.project_loader import ProjectTemplate
from zetsubou.utils.common import Version


def is_second_newer_toolchain(first:Toolchain, second:Toolchain) -> bool:
    if first.CompilerFamily == ECompilerFamily.MSVC:
        if first.ide_version < second.ide_version:
            return True

    return first.version < second.version


def get_toolchains_per_platform(project_template: ProjectTemplate) -> Dict[str, List[str]]:
    result = {}

    for platform in project_template.platforms:
        compatible_toolchains = set()

        for tool_filter in platform.toolchains:
            latest_by_compiler : Toolchain = None
            latest_by_toolset : Toolchain = None

            filter_by_toolset = isinstance(tool_filter, IWindowsToolchain)

            for tool_entry in project_template.toolchains:
                if tool_entry.CompilerFamily != tool_filter.family:
                    continue # Family mismatch

                if tool_filter.version == EVersionSelector.latest:
                    if latest_by_compiler is None:
                        latest_by_compiler = tool_entry
                    elif is_second_newer_toolchain(latest_by_compiler, tool_entry):
                        latest_by_compiler = tool_entry
                    continue

                elif tool_filter.version != EVersionSelector.all and tool_entry.version not in tool_filter.version:
                    continue # Version mismatch

                if filter_by_toolset:
                    if tool_filter.toolset == EVersionSelector.latest:
                        if latest_by_toolset is None:
                            latest_by_toolset = tool_entry
                        elif latest_by_toolset.Toolset < tool_entry.Toolset:
                            latest_by_toolset = tool_entry
                        continue

                    elif tool_filter.toolset != EVersionSelector.all and tool_entry.Toolset not in tool_filter.toolset:
                        continue # Version mismatch

                compatible_toolchains.add(tool_entry)

            if latest_by_compiler is not None:
                compatible_toolchains.add(latest_by_compiler)

            if latest_by_toolset is not None:
                compatible_toolchains.add(latest_by_toolset)

        # --

        result[platform.platform] = [ t.name for t in compatible_toolchains]

    return result



import winreg
import os, json
from typing import List, Optional, Any
from dataclasses import dataclass
from zetsubou.project.model.platform import EArch

from zetsubou.project.model.toolchain import ECompilerFamily, Toolchain
from zetsubou.system.windows import vswhere
from zetsubou.utils import logger
from zetsubou.utils.common import Version, get_env_path, get_files, get_subdirs
from zetsubou.utils.subprocess import call_process


def is_current_host(host: str, host_arch: EArch) -> bool:
    # x64
    if host.endswith('64'):
        if host_arch == EArch.x64:
            return True
    # x86 (32-bit)
    else:
        if host_arch != EArch.x64:
            return True
    return False


@dataclass
class Toolset:
    full_path: str
    identifier: str

    def __lt__(self, other):
        return self.identifier > other.identifier


def find_toolsets(msvc_path : str) -> List[Toolset]:
    toolsets = []
    search_subpath = 'Msbuild\\Microsoft\\VC'
    search_path = os.path.join(msvc_path, search_subpath)
    if os.path.exists(search_path):
        for root, _, files in os.walk(search_path):
            if 'Toolset.props' in files:
                identifier = os.path.basename(root)
                if identifier.startswith('v'):
                    toolsets.append(Toolset(full_path=root, identifier=identifier))

    return toolsets


def get_toolset_ver(toolset:Toolset):
    if toolset.identifier.startswith('v'):
        return int(toolset.identifier[1:])
    return 0


# Value defined in _MSVC_VER
def get_cl_exe_compiler_version(cl_exe_path:str):
    out, err = call_process([cl_exe_path], capture=True, realtime=False)
    if err is not None:
        return None

    # Strip prefix
    version = ''
    for i in range(len(out)):
        if out[i:i+1].isdigit():
            version = out[i:]
            break

    # Strip suffix
    version = version.split()[0]

    return version


# Filters out legacy versions of toolsets, leaving only the latest one
def filter_legacy_toolsets(msvc_path : str, toolsets: List[Toolset]) -> List[Toolset]:
    msvc_path = os.path.join(msvc_path, 'Msbuild\\Microsoft\\VC\\')
    selected_version = ''
    result = []

    # Filter out legacy toolsets
    for tool in toolsets:
        if tool.full_path.startswith(msvc_path):
            tmp_path = tool.full_path[len(msvc_path):]
            first_slash_idx = tmp_path.find('\\')
            tmp_path = tmp_path[:first_slash_idx]
            if tmp_path > selected_version:
                selected_version = tmp_path
    for tool in toolsets:
        if selected_version in tool.full_path:
            result.append(tool)

    return result


def get_registry_key(storage : str, key : str, entry : str) -> Optional[Any]:
    try:
        hkey = winreg.OpenKey(storage, key)
    except:
        return None
    else:
        try:
            value, _ = winreg.QueryValueEx(hkey, entry)
            return value
        except:
            return None
        finally:
            winreg.CloseKey(hkey)


@dataclass
class WindowsSDKPaths:
    version : str
    base_path : str
    include : str
    lib : str
    bin : str


def discover_win10_sdk():
    storages = [winreg.HKEY_LOCAL_MACHINE, winreg.HKEY_CURRENT_USER]
    key_paths = ['SOFTWARE\\Wow6432Node', 'SOFTWARE']

    def is_valid_sdk_ver(sdk_ver:str) -> bool:
        return sdk_ver.startswith('10.')

    for storage in storages:
        for key in key_paths:

            key_path = f'{key}\\Microsoft\\Microsoft SDKs\\Windows\\v10.0'
            install_dir = get_registry_key(storage, key_path, 'InstallationFolder')

            if install_dir and os.path.isdir(install_dir):
                include_dir = os.path.join(install_dir, 'include')

                sdk_versions = filter(is_valid_sdk_ver, os.listdir(include_dir))
                latest_sdk = max(sdk_versions)

                logger.Verbose(f"Possible Win10SDK {install_dir}\nlatest: {latest_sdk}\nall:{ ', '.join(sdk_versions) }")

                if os.path.isdir(os.path.join(include_dir, latest_sdk)) and latest_sdk.startswith('10.'):
                    windows_h = os.path.join(include_dir, latest_sdk, 'um', 'Windows.h')
                    if os.path.isfile(windows_h):
                        sdk = WindowsSDKPaths(
                            version=latest_sdk,
                            base_path=install_dir,
                            include=os.path.join(install_dir, 'Include', latest_sdk),
                            lib=os.path.join(install_dir, 'Lib', latest_sdk),
                            bin=os.path.join(install_dir, 'bin', latest_sdk),
                        )
                        return sdk
    return None


def discover_msvc(host_arch: EArch, vs_instances) -> List[Toolchain]:
    toolchains : List[Toolchain] = []

    default_vc_path = 'VC\\Tools\\MSVC'
    default_bin_path = 'bin'
    default_lib_path = 'lib'
    default_include_paths = ['include', 'atlmfc\\include']
    default_compiler_msvc = 'cl.exe'
    default_linker_msvc = 'link.exe'
    default_librarian_msvc = 'lib.exe'

    additional_include_dirs = [
        'VC\\Auxiliary\\VS\\include',
        'VC\\Auxiliary\\VS\\UnitTest\\include'
    ]

    default_extra_files = [
        'c1.dll',
        'c2.dll',
        'c1xx.dll',
        '1033\\clui.dll',
        'msobj140.dll',
        'mspdb140.dll',
        'mspdbsrv.exe',
        'mspdbcore.dll',
        'mspft140.dll',
        'msvcp140.dll',
        'vcruntime140.dll',
        'vccorlib140.dll'
    ]

    default_link_libraries = [
        'kernel32'
        'user32'
        'gdi32'
        'winspool'
        'comdlg32'
        'advapi32'
        'shell32'
        'ole32'
        'oleaut32'
        'uuid'
        'odbc32'
        'odbccp32'
        'delayimp'
    ]

    for vs_itr in vs_instances:
        installation_path = vs_itr['installationPath']
        vs_ide_version = vs_itr['installationVersion']
        vc_tools = os.path.join(installation_path, default_vc_path)
        vs_version = vs_itr['catalog']['productLineVersion']

        available_toolsets = find_toolsets(installation_path)

        if os.path.exists(vc_tools):

            vc_versions = get_subdirs(vc_tools)
            vc_ver = max(vc_versions)

            # for vc_ver in vc_versions:
            vc_ver_basepath = os.path.join(vc_tools, vc_ver)
            vc_ver_binpath = os.path.join(vc_ver_basepath, default_bin_path)
            vc_ver_libpath = os.path.join(vc_ver_basepath, default_lib_path)

            for host in get_subdirs(vc_ver_binpath):
                if is_current_host(host, host_arch):
                    for arch in get_subdirs(os.path.join(vc_ver_binpath, host)):

                        msvc_name = vc_ver.replace('.','_')
                        arch_bin_fullpath = os.path.join(vc_ver_binpath, host, arch)
                        arch_lib_fullpath = os.path.join(vc_ver_libpath, arch)

                        visited_toolsets = set()

                        for toolset in available_toolsets:
                            if arch in toolset.full_path and toolset.identifier not in visited_toolsets:
                                visited_toolsets.add(toolset.identifier)
                                toolchain_name = f'{ECompilerFamily.MSVC.name}_{arch}_vs{vs_version}_v{msvc_name}_{toolset.identifier}'

                                includes = []
                                includes += [ os.path.join(vc_ver_basepath, inc) for inc in default_include_paths ]
                                includes += [ os.path.join(installation_path, inc) for inc in additional_include_dirs ]

                                # TODO: maybe this should be separate for compiler, librarian and linker?
                                path_envs = []
                                path_envs += [ arch_bin_fullpath ]

                                machine_linker_options = f'/MACHINE:{arch} '

                                compiler = Toolchain.create(
                                    toolchain_name,
                                    os.path.join(arch_bin_fullpath, default_compiler_msvc),
                                    os.path.join(arch_bin_fullpath, default_librarian_msvc),
                                    os.path.join(arch_bin_fullpath, default_linker_msvc),

                                    arch=arch,

                                    version=Version(vc_ver),
                                    ide_version=Version(vs_ide_version),
                                    vs_install_path=installation_path,

                                    Toolset=toolset.identifier,

                                    PathEnv=path_envs,

                                    LinkerOptions=machine_linker_options,

                                    LinkerPaths=[
                                        arch_lib_fullpath,
                                        arch_bin_fullpath,
                                    ],

                                    ExtraFiles = default_extra_files,

                                    RequiredFiles=[],
                                    IncludeDirectories=includes,
                                    Defines=[]
                                )
                                toolchains.append(compiler)

    return toolchains


def get_clang_version(clang_exe_path: str) -> Optional[str]:
    out, err = call_process([clang_exe_path, '--version'], capture=True, realtime=False)
    if err is not None:
        return None

    version_prefix = 'clang version '
    version = out[len(version_prefix):]
    version = version.split()[0]

    return version


def discover_clang(host_arch:EArch, vs_instances) -> List[Toolchain]:
    # We need to find visual studio installations anyway
    # But how to handle standalone LLVM? Ignore it? Or match with what version?
    # All of them? The latest? Per toolset?

    # There is a thing I cannot find documentation about
    # Seems the Clang toolset exists in two versions per visual studio installation
    #   egx. 2022 has v170 and v160 folders with Clang
    #    and 2019 has v160 and v150
    # We don't want th
    # toolsets = filter_legacy_toolsets(installation_path, toolsets)

    # Find VS installations
    # Find Clang from visual studios

    # Find standalone LLVM installation
    # Match with vs toolsets

    # clang.exe will work as an compiler, librarian and linker
    # but we need link.exe to be present in path when calling the compiler
    # how on earth would we do that in fastbuild..?
    # Compiler in fastbuild has environment option!

    # TODO: We need to pass Environment from Compiler to DLLs, Libraries and Execs!
    # This will set correct PATH with linker.exe
    toolchains : List[Toolchain] = []

    default_msvc_path = 'VC\\Tools\\MSVC'
    default_llvm_path = 'VC\\Tools\\Llvm'
    default_bin_path = 'bin'
    default_lib_path = 'lib'
    default_llvm_include_path = 'include'
    default_msvc_include_paths = ['include', 'atlmfc\\include']
    default_compiler_clang = 'clang.exe'
    default_librarian_clang = 'llvm-ar.exe'
    default_compiler_msvc = 'cl.exe'
    default_linker_msvc = 'link.exe'

    additional_include_dirs = [
        'VC\\Auxiliary\\VS\\include',
        'VC\\Auxiliary\\VS\\UnitTest\\include'
    ]

    default_extra_files = [
        'c1.dll',
        'c2.dll',
        'c1xx.dll',
        '1033\\clui.dll',
        'msobj140.dll',
        'mspdb140.dll',
        'mspdbsrv.exe',
        'mspdbcore.dll',
        'mspft140.dll',
        'msvcp140.dll',
        'vcruntime140.dll',
        'vccorlib140.dll'
    ]

    standalone_clangs = []

    def try_create_clang_toolchain(clang_exe_path, msvc_ver_basepath, msvc_host, msvc_arch, vs_version, toolset:Toolset) -> Optional[Toolchain]:
        clang_ver = get_clang_version(clang_exe_path)
        if clang_ver is None:
            return None

        vc_ver_binpath = os.path.join(msvc_ver_basepath, default_bin_path)
        vc_ver_libpath = os.path.join(msvc_ver_basepath, default_lib_path)
        msvc_arch_bin_fullpath = os.path.join(vc_ver_binpath, msvc_host, msvc_arch)
        msvc_arch_lib_fullpath = os.path.join(vc_ver_libpath, msvc_arch)
        msvc_linker_exe = os.path.join(msvc_arch_bin_fullpath, default_linker_msvc)

        msvc_ver = get_cl_exe_compiler_version(os.path.join(msvc_arch_bin_fullpath, default_compiler_msvc))
        if msvc_ver is None:
            return None

        clang_ver_name = clang_ver.replace('.','_')

        clang_bin_dir = os.path.dirname(clang_exe_path)
        clang_base_dir = os.path.dirname(clang_bin_dir)
        clang_lib_dir = os.path.join(clang_base_dir, default_lib_path)
        clang_include_dir = os.path.join(clang_base_dir, default_llvm_include_path)
        clang_librarian_exe = os.path.join(clang_bin_dir, default_librarian_clang)

        compiler_name = f"{ECompilerFamily.CLANG.name}_{msvc_arch}_vs{vs_version}_v{clang_ver_name}_{toolset.identifier}"

        includes = [ clang_include_dir ]
        includes += [ os.path.join(msvc_ver_basepath, inc) for inc in default_msvc_include_paths ]
        includes += [ os.path.join(installation_path, inc) for inc in additional_include_dirs ]

        # TODO: maybe this should be separate for compiler, librarian and linker?
        path_envs = []
        path_envs += [ clang_bin_dir, msvc_arch_bin_fullpath ]

        machine_linker_options = f'/MACHINE:{msvc_arch} '

        sem_msvc_ver = Version(msvc_ver)
        msvc_compat_options = f"-fms-compatibility-version={sem_msvc_ver.major}{sem_msvc_ver.minor} "

        return Toolchain.create(
            compiler_name,
            clang_exe_path,
            clang_librarian_exe,
            msvc_linker_exe,

            CompilerOptions = msvc_compat_options,
            LinkerOptions = machine_linker_options,

            arch = msvc_arch,
            version = Version(clang_ver),
            ide_version = Version(vs_ide_version),
            vs_install_path = installation_path,

            Toolset = toolset.identifier,
            PathEnv = path_envs,
            ExtraFiles = default_extra_files,

            LinkerPaths = [
                clang_lib_dir,
                clang_bin_dir,
                msvc_arch_lib_fullpath,
                msvc_arch_bin_fullpath,
            ],

            RequiredFiles=[],
            IncludeDirectories=includes,
            Defines=[]
        )

    env_path = get_env_path()
    for path in env_path:
        for file in get_files(path): 
            if file.find(default_compiler_clang) != -1:
                clang_path = os.path.join(path, file)
                standalone_clangs.append(clang_path)

    for vs_itr in vs_instances:
        installation_path = vs_itr['installationPath']
        vs_ide_version = vs_itr['installationVersion']
        msvc_tools = os.path.join(installation_path, default_msvc_path)
        llvm_tools = os.path.join(installation_path, default_llvm_path)
        vs_version = vs_itr['catalog']['productLineVersion']

        toolsets = find_toolsets(installation_path)
        latest_toolset = max(toolsets, key=get_toolset_ver)

        logger.Verbose(f"For VS {installation_path} picked {latest_toolset} from found toolsets:\n{[t.identifier for t in toolsets]}")

        msvc_versions = get_subdirs(msvc_tools)
        latest_msvc = max(msvc_versions)

        # for vc_ver in vc_versions:
        msvc_ver_basepath = os.path.join(msvc_tools, latest_msvc)
        vc_ver_binpath = os.path.join(msvc_ver_basepath, default_bin_path)

        if os.path.exists(msvc_tools):

            for msvc_host in get_subdirs(vc_ver_binpath):
                if is_current_host(msvc_host, host_arch):
                    for msvc_arch in get_subdirs(os.path.join(vc_ver_binpath, msvc_host)):

                        # Standalone Clang
                        for clang_entry in standalone_clangs:
                            compiler = try_create_clang_toolchain(clang_entry, msvc_ver_basepath, msvc_host, msvc_arch, vs_version, latest_toolset)
                            if compiler is None:
                                continue

                            toolchains.append(compiler)

                        # Bundled Clang
                        if os.path.exists(llvm_tools):
                            for host in get_subdirs(llvm_tools):
                                if is_current_host(host, host_arch):
                                    if host == default_bin_path:
                                        host = ''

                                    clang_dir = os.path.join(llvm_tools, host, default_bin_path)
                                    clang_exe = os.path.join(clang_dir, default_compiler_clang)

                                    compiler = try_create_clang_toolchain(clang_exe, msvc_ver_basepath, msvc_host, msvc_arch, vs_version, latest_toolset)
                                    if compiler is None:
                                        continue

                                    toolchains.append(compiler)

    return toolchains


def discover_toolchain_list(host_arch: EArch) -> List[Toolchain] :
    vswhere_path = None

    try:
        vswhere_path = vswhere.find_vswhere()
    except Exception:
        pass

    if vswhere_path is None:
        logger.Warning("Unable to find vswhere executable, VS toolchains not discovered")
        return []

    vs_instances = json.loads(vswhere.call_vswhere(vswhere_path))
    if vs_instances is None:
        logger.Warning("vswhere found no VS installations")
        return []

    toolchains : List[Toolchain] = []
    toolchains += discover_msvc(host_arch, vs_instances)
    toolchains += discover_clang(host_arch, vs_instances)

    win10_sdk = discover_win10_sdk()

    if win10_sdk is not None:
        logger.Verbose(f"Found Win10SDK in {win10_sdk.base_path}")
        for toolchain in toolchains:
            # This should work for clang/clangCL too
            #if toolchain.CompilerFamily == ECompilerFamily.MSVC:
            toolchain.LinkerPaths.append(f'{win10_sdk.lib}\\um\\{toolchain.arch.lower()}')
            toolchain.LinkerPaths.append(f'{win10_sdk.lib}\\ucrt\\{toolchain.arch.lower()}')
            toolchain.IncludeDirectories.append(f'{win10_sdk.include}\\shared')
            toolchain.IncludeDirectories.append(f'{win10_sdk.include}\\um')
            toolchain.IncludeDirectories.append(f'{win10_sdk.include}\\ucrt')

    return toolchains



from copy import copy
from zetsubou.commands.base_command import CommandContext
from typing import List
from zetsubou.project.model.configuration import Configuration
from zetsubou.project.model.platform import EArch, ESystem
from zetsubou.project.model.runtime_library import ERuntimeLibrary

from zetsubou.project.model.toolchain import ECompilerFamily, Toolchain


def platform_to_conan(context: CommandContext, plat : str) -> List[str]:
    return [
        '-s', f'os={plat}'
    ]


def toolchain_to_conan(context: CommandContext, toolchain: Toolchain) -> List[str]:
    if toolchain is not None:

        arch = {
            EArch.x86.name : 'x86',
            EArch.x64.name : 'x86_64'
        }.get(toolchain.arch)

        compiler = {
            ECompilerFamily.MSVC.name  : 'Visual Studio',
            ECompilerFamily.CLANG.name : 'clang',
            ECompilerFamily.GCC.name   : 'gcc'
        }.get(toolchain.CompilerFamily.name)

        result = [
            '-s', f'arch={arch}',
            '-s', f'compiler={compiler}',
        ]

        for path in toolchain.PathEnv:
            result.extend([
                '-e', f'PATH=[{path}]'
            ])

        if context.host_system == ESystem.Windows and toolchain.Toolset is not None:
            if toolchain.CompilerFamily == ECompilerFamily.MSVC:
                result.extend([
                    '-s', f'compiler.toolset={toolchain.Toolset}'
                ])
            elif toolchain.CompilerFamily == ECompilerFamily.CLANG:
                result.extend([
                    '-s', f'compiler.runtime_version={toolchain.Toolset}'
                ])

        if toolchain.CompilerFamily == ECompilerFamily.MSVC:
            result.extend([
                '-s', f'compiler.version={toolchain.ide_version.major}'
            ])
        else:
            # Remove path, conan is interested only in Major.Minor at best
            ver_no_path = copy(toolchain.version)
            ver_no_path.path = 0

            result.extend([
                '-s', f'compiler.version={ver_no_path}'
            ])

        return result

    return []


def config_base_to_conan(context: CommandContext, config: Configuration, toolchain:Toolchain) -> List[str]:
    if config is not None and toolchain is not None:

        build_type = {
            'DEBUG' : 'Debug',
            'RELEASE' : 'Release',
            'RELEASE_WITH_DEBUG_INFO' : 'RelWithDebInfo'
        }.get(config.base_configuration.name)

        result = [
            '-s', f'build_type={build_type}',
        ]

        if context.host_system == ESystem.Windows:

            if toolchain.CompilerFamily == ECompilerFamily.MSVC:
                runtime_library = {
                    ERuntimeLibrary.DYNAMIC_DEBUG.name   : 'MDd',
                    ERuntimeLibrary.DYNAMIC_RELEASE.name : 'MD',
                    ERuntimeLibrary.STATIC_DEBUG.name    : 'MTd',
                    ERuntimeLibrary.STATIC_RELEASE.name : 'MT',
                }.get(config.runtime_library.name)

                result.extend([
                    '-s', f'compiler.runtime={runtime_library}'
                ])
            elif toolchain.CompilerFamily == ECompilerFamily.CLANG:
                runtime, runtime_type = {
                    ERuntimeLibrary.DYNAMIC_DEBUG.name   : ('dynamic', 'Debug'),
                    ERuntimeLibrary.DYNAMIC_RELEASE.name : ('dynamic', 'Release'),
                    ERuntimeLibrary.STATIC_DEBUG.name    : ('static',  'Debug'),
                    ERuntimeLibrary.STATIC_RELEASE.name  : ('static',  'Release'),
                }.get(config.runtime_library.name)

                result.extend([
                    '-s', f'compiler.runtime={runtime}',
                    '-s', f'compiler.runtime_type={runtime_type}',
                ])

        return result

    return []



from dataclasses import dataclass
from fs.base import FS
from zetsubou.project.model.platform import ESystem, EArch


@dataclass
class BaseContext:
    host_system: ESystem = 0
    host_arch: EArch = 0
    fs_root: str = ''
    fs_venv: str = ''
    project_fs: FS = ''
    project_file: str = ''



from zetsubou.project.model.platform import EArch
from zetsubou.utils.error import ProjectError

def arch_to_msvc_platform(arch : EArch):
    lookup = {
        EArch.x64 : 'x64',
        EArch.x86 : 'Win32'
    }

    if arch not in lookup:
        raise ProjectError(f"Arch '{arch}' is not supported for MSVC!")

    return lookup.get(arch)



from dataclasses import dataclass
from typing import List, Optional
import importlib_resources as resources

import fs
import os
from fs.base import FS
from zetsubou.project.model.config_string import EDefaultConfigSlots
from zetsubou.project.model.configuration import Configuration
from zetsubou.project.model.kind import ETargetKind
from zetsubou.project.model.platform import Platform
from zetsubou.project.model.project_template import Project
from zetsubou.project.model.rule import Rule
from zetsubou.project.model.target import Dependencies, Target, TargetReference
from zetsubou.project.model.toolchain import Toolchain
from zetsubou.utils import logger, yaml_loader
from zetsubou.utils.common import null_or_empty
from zetsubou.utils.error import ProjectError

import bentoudev.dataclass.base as base


def foreach(list, fn):
    if list is None:
        return
    if len(list) == 0:
        return
    for element in list:
        fn(element)


def find_target(targets: List[Target], name: str) -> Optional[Target]:
    for target in targets:
        if target.target == name:
            return target
    return None


class ValidationContext:
    error_format: base.EErrorFormat


def resolve_target_references(ctx:ValidationContext, targets: List[Target], deps: Optional[List[TargetReference]]) -> List[TargetReference]:
    failed_refs = []
    if deps:
        for ref in deps:
            found_target = find_target(targets, ref.name)
            if found_target is None:
                failed_refs.append(ref)
            else:
                ref.target = found_target
    return failed_refs


def validate_duplicates(ctx:ValidationContext, targets: List[Target]):
    visited = dict()
    for target in targets:
        if target.target in visited:
            tmp_msg = target.format_field_message('target', 'error', f"Duplicated target name '{target.target}'!", ctx.error_format)
            err_msg = f"{tmp_msg}\n'{visited[target.target].format_field_message('target', '', 'Second', ctx.error_format)}'"
            raise ProjectError(err_msg)
        else:
            visited[target.target] = target


def validate_cyclic_dependency(ctx:ValidationContext, targets: List[Target]):
    visited = dict()

    def visit_target(in_target):
        nonlocal visited

        target = in_target

        if isinstance(in_target, TargetReference):
            target = in_target.target

        if target.target in visited:
            target_path = []
            msg = []

            for path_element, path_ref in visited.items():
                target_path.append(path_element)
                target_path.append(' -> ')

                if isinstance(path_ref, TargetReference):
                    msg.append(path_ref.format_message('', '\nReferenced here', ctx.error_format))
                    msg.append('\n')

            target_path.append(f'{target.target}')

            path_msg = ''.join(target_path)
            ref_msg = ''.join(msg)

            main_msg = in_target.format_message('error', f"Cyclic dependency to target '{target.target}' through '{path_msg}'", ctx.error_format)
            err_msg = f'{main_msg}\n{ref_msg}'
            raise ProjectError(err_msg)

        def visit_dependencies(t: Target, d: Dependencies):
            foreach(d.interface, visit_target)
            foreach(d.public, visit_target)
            foreach(d.private, visit_target)

        visited[target.target] = in_target

        if target.dependencies is not None:
            visit_dependencies(target, target.dependencies)

        if target.filters is not None:
            for f in target.filters:
                if f.dependencies is not None:
                    visit_dependencies(target, f.dependencies)

        del visited[target.target]

    for target in targets:
        visit_target(target)


def resolve_target_dependencies(ctx:ValidationContext, target_name: str, dependencies, targets: List[Target]):
    if dependencies is not None:
        for name, dep_list in {
            'interface': dependencies.interface,
            'public': dependencies.public,
            'private': dependencies.private
        }.items():

            failed_refs = resolve_target_references(ctx, targets, dep_list)

            if len(failed_refs) > 0:
                msg = ''
                for failed_ref in failed_refs:
                    msg += failed_ref.format_message('error', f"\nUnknown target '{failed_ref.name}'", base.EErrorFormat.Pretty)

                raise ProjectError(
                    f'Reference to undefined targets in \'{name}\' dependencies of target \'{target_name}\'{msg}')


def resolve_targets_dependencies(ctx:ValidationContext, targets: List[Target]):
    for target in targets:
        resolve_target_dependencies(ctx, target.target, target.dependencies, targets)
        for f in target.filters:
            resolve_target_dependencies(ctx, target.target, f.dependencies, targets)


def validate_targets_not_empty(ctx:ValidationContext, targets: List[Target]):
    if (len(targets)) == 0:
        raise ProjectError("Cannot load project with no targets!")


def validate_unresolved_targets(ctx:ValidationContext, targets: List[Target]):
    validate_targets_not_empty(ctx, targets)


def validate_targets(ctx:ValidationContext, targets : List[Target]):
    validate_unresolved_targets(ctx, targets)
    resolve_targets_dependencies(ctx, targets)
    validate_duplicates(ctx, targets)
    validate_cyclic_dependency(ctx, targets)


##########################################################################


@dataclass
class ProjectTemplate:
    project: Project
    platforms: List[Platform]
    toolchains: List[Toolchain]
    configurations: List[Configuration]
    rules: List[Rule]
    targets: List[Target]

    def find_target(self, name:str):
        return find_target(self.targets, name)

    def find_platform(self, name : str):
        for plat in self.platforms:
            if plat.platform == name:
                return plat
        return None

    def find_config(self, name : str):
        for conf in self.configurations:
            if conf.configuration == name:
                return conf
        return None

    def find_toolchain(self, name: str):
        for tool in self.toolchains:
            if name == tool.name:
                return tool
        return None

    def list_configurations(self):
        result = []
        for conf in self.configurations:
            result.append(conf.configuration)
        return result

    def list_platforms(self):
        result = []
        for plat in self.platforms:
            result.append(plat.platform)
        return result

    def list_toolchain(self):
        result = []
        for tool in self.toolchains:
            result.append(tool.name)
        return result

    def slot_values(self):
        result = {
            EDefaultConfigSlots.platform.name : self.list_platforms(),
            EDefaultConfigSlots.configuration.name : self.list_configurations(),
            EDefaultConfigSlots.toolchain.name : self.list_toolchain()
        }

        for option in self.project.options:
            result[option.name] = option.values

        return result

    @staticmethod
    def compile_target_variant_name(kind: ETargetKind, name: str, config_string: str = ''):
        parts = []

        if kind != ETargetKind.INVALID:
            parts.append(f'{name}_{kind.name}')
        else:
            parts.append(name)

        if not null_or_empty(config_string):
            parts.append(config_string)

        return '-'.join(parts)

    @staticmethod
    def decompose_variant_name(variant:str, template_str:str):
        parts = variant.split('-')
        target = parts[0]
        config_parts = parts[1:]

        slots = template_str.split('-')

        if len(config_parts) != len(slots):
            return None, None

        result = {}
        for slot, val in zip(slots, config_parts):
            slot_name = slot.strip('}{')
            result[slot_name] = val

        return target, result


def resolve_venv(project_fs : FS, proj : Project, proj_path : str):
    if proj.config.venv == '':
        env_dir = ''
        for path in project_fs.walk.files(filter=['activate.*']):
            env_dir = fs.path.dirname(path)
            if env_dir.startswith('/'):
                env_dir = env_dir[1:]

        if env_dir == '':
            proj_yml_path = proj.get_loaded_from_file()
            logger.CriticalError('Virtual environment not found! \n'
                f" Please, either set path to config.venv in {proj_yml_path}, run 'zetsubou install' command, or place activate.bat in one of the project subdirectories ")
            return None

        return env_dir

    else:
        venv = os.path.abspath(os.path.join(proj_path, proj.config.venv))

        if os.path.exists(venv) and os.path.isdir(venv):
            def walk_for_env(search_root : str) -> Optional[str]:
                for root, _, files in os.walk(search_root):
                    for file in files:
                        if file.find('activate.') != -1:
                            return root
                return None

            venv = walk_for_env(venv)

        if venv is None:
            raise ProjectError(
                f"Virtual environment not found! Path '{proj.config.venv}' doesnt exist or is not a directory!\n"
                f'{proj.config.format_field_message("venv")}')
        else:
            return venv


def load_dataclass_list(clazz: type, obj_ref_list: Optional[List[str]], proj_dir : str, project_fs : FS, loader, local_types):
    result = []

    if obj_ref_list is not None and len(obj_ref_list) > 0:

        for obj_ref in obj_ref_list:
            obj_path = fs.path.join(proj_dir, obj_ref)

            if not project_fs.exists(obj_path) or not project_fs.isfile(obj_path):
                raise ProjectError(f"Unknown to locate file '{obj_path}'")

            with project_fs.open(obj_path, 'r', encoding='utf-8') as obj_file:
                obj_templ = loader.load_dataclass(clazz, obj_path, obj_file.read(), local_types)

                if obj_templ is not None:
                    if not base.is_loaded_from_file(obj_templ):
                        raise ProjectError(f"Unable to load class '{type(obj_templ)}' from file, missing attribute 'loadable_from_file'!")

                    if logger.IsVisible(logger.ELogLevel.Verbose):
                        logger.Success(f"Loaded [{clazz.__name__}] '{obj_path}'")

                    obj_templ.set_loaded_from_file(obj_path)
                    result.append(obj_templ)

    return result


def load_bundled_rules(loader, local_types) -> List[Rule]:
    result = []
    for res_path in [ 'MsvcRules.yml', 'ClangRules.yml' ]:
        res_content = resources.files('zetsubou.data.rules').joinpath(res_path).read_text()
        obj_templ = loader.load_dataclass(Rule, res_path, res_content, local_types)

        if obj_templ is not None:
            if not base.is_loaded_from_file(obj_templ):
                raise ProjectError(f"Unable to load class '{type(obj_templ)}' from file, missing attribute 'loadable_from_file'!")

            if logger.IsVisible(logger.ELogLevel.Verbose):
                logger.Success(f"Loaded [Rule] '{res_path}'")

            obj_templ.set_loaded_from_file(res_path)
            result.append(obj_templ)
    return result


def load_project_template(project_fs: FS, fs_root : str, filename: str, proj_file_content: str, *, loader=yaml_loader.YamlDataclassLoader()) -> Optional[ProjectTemplate]:
    try:
        local_types = base.get_types_from_modules([__name__, 'zetsubou.project.model.target'])

        proj_dir = os.path.dirname(filename)
        proj_templ: Project = loader.load_dataclass(Project, filename, proj_file_content, local_types)
        if proj_templ is None:
            return

        proj_templ.set_loaded_from_file(filename)

        # TODO validate config string in proj

        # First load targets, resolve references and report undefined ones, validate for cyclic dependency
        targets = load_dataclass_list(Target, proj_templ.targets, proj_dir, project_fs, loader, local_types)

        proj = ProjectTemplate(
            project=proj_templ,
            platforms=load_dataclass_list(Platform, proj_templ.config.platforms, proj_dir, project_fs, loader, local_types),
            toolchains=[],
            configurations=load_dataclass_list(Configuration, proj_templ.config.configurations, proj_dir, project_fs, loader, local_types),
            rules=load_dataclass_list(Rule, proj_templ.config.rules, proj_dir, project_fs, loader, local_types),
            targets=targets,
        )

        proj.rules += load_bundled_rules(loader, local_types)

        return proj

    except base.DataclassLoadError as error:
        logger.Error(error)
        logger.CriticalError(f'Failed to load project \'{filename}\'')
        return None

    except ProjectError as error:
        logger.Error(error)
        logger.CriticalError(f'Failed to load project \'{filename}\'')
        return None

    except Exception as error:
        logger.Exception(error)
        logger.CriticalError(f'Failed to load project  \'{filename}\'')
        return None


def load_project_from_file(project_fs: FS, fs_root : str, filename: str) -> Optional[ProjectTemplate]:
    if not project_fs.exists(filename):
        logger.CriticalError(f"Unable to locate file '{filename}'")
        return None
    with project_fs.open(filename, 'r', encoding='utf-8') as proj_file:
        return load_project_template(project_fs, fs_root, filename, proj_file.read())



import time

from zetsubou.utils import logger
from zetsubou.utils.error_codes import EErrorCode, ReturnErrorcode

def execute_stage(func, succ_msg: str, err_code: EErrorCode):
    start_time = time.time()
    result = func()
    end_time = time.time()

    if result is not None and result is not False:
        logger.Success(f'{succ_msg} - {end_time - start_time:.2f}sec')
        return result
    else:
        raise ReturnErrorcode(err_code)



from conans.model import Generator
from conans import ConanFile
import os


def emit_list(lst):
    return '\n       - '.join(lst)


def emit_property(out, prop, tmpl):
    if prop is not None and len(prop) > 0:
        out.append(tmpl.format(data = emit_list(prop)))


def fix_no_ext(libs, paths):
    results = []
    for path in paths:
        for _, _, filenames in os.walk(path):
            for file in filenames:
                for lib in libs:
                    if file.find(lib) != -1:
                        results.append(os.path.basename(file))
    return results


class zetsubou_multi(Generator):
    @property
    def filename(self):
        pass

    @property
    def content(self):
        result = {}

        for dep_name, dep_cpp_info in self.deps_build_info.dependencies:
            sections = []
            emit_property(sections, dep_cpp_info.include_paths,
                '    includes:\n'
                '       interface:\n'
                '       - {data}\n')

            emit_property(sections, dep_cpp_info.defines,
                '    defines:\n'
                '       interface:\n'
                '       - {data}\n')

            emit_property(sections, dep_cpp_info.cppflags,
                '    compiler_flags:\n'
                '       interface:\n'
                '       - {data}\n')

            emit_property(sections, fix_no_ext(dep_cpp_info.libs, dep_cpp_info.lib_paths),
                '    link_libraries:\n'
                '       interface:\n'
                '       - {data}\n')

            emit_property(sections, dep_cpp_info.lib_paths,
                '    linker_paths:\n'
                '       interface:\n'
                '       - {data}\n')

            emit_property(sections, dep_cpp_info.sharedlinkflags + dep_cpp_info.exelinkflags,
                '    linker_flags:\n'
                '       interface:\n'
                '       - {data}\n')

            dep_name = dep_name.replace("-", "_")
            result[f'{dep_name}.part.yml'] = ''.join(sections)

        # result['conan_multi.txt'] = '\n'.join([ name for name, _ in self.deps_build_info.dependencies ])

        return result


class ZetsubouGeneratorPackage(ConanFile):
    name = "zetsubougen"
    version = "0.6.1"
    url = "https://github.com/..."
    license = "MIT"
    description = 'Zetsubou yml files generator'

    def build(self):
        pass

    def package_info(self):
        self.cpp_info.includedirs = []
        self.cpp_info.libdirs = []
        self.cpp_info.bindirs = []
